with Controller;
with Common;
with Results;
with Ada.Text_IO;

package body Prime is

   type Prime_Index is range 0 .. Common.Capacity;

   --  the prime numbers known to this partition
   Local_Prime_Table   : array (Prime_Index) of Natural;

   --  the index of the last prime in Local_Prime_Table
   Last_Prime_Index    : Prime_Index := 0;

   --  the index of the prime to test againt the current number
   Current_Prime_Index : Prime_Index := 1;

   --  remember which number we are testing in order to reset
   --+ Current_Prime_Index when a new number is tested, so
   --+ that we test it against all values in Local_Prime_Table
   Number_To_Test      : Natural := 0;

   --  attribute 'Partition_ID of unit U returns an integer
   --+ which is the unique id that identifies the partition
   --+ on which unit U is located
   Local_Partition_ID  : Common.Partition_ID := Prime'Partition_ID;

   --  distributed objects for the current and next partition
   --+ (note that Pool_Access is a class-wide type)
   Self_Pool : aliased New_Pool_Type;
   Next_Pool : Controller.Pool_Access := null;

   --  register the distributed object of the current partition
   --+ to the Controller, using a dynamically bound call, which
   --+ is dispatched according to the value of its controlling
   --+ parameter (Self_Pool'access) of access-to-class-wide type
   --+
   --+ this source line is executed during partition elaboration
   --+ its only purpose is to identify the calling partition to
   --+ the Controller register (RCI)
   --+ (the configuration file will allocate Controller to a
   --+  specific partition)
   Static_ID : Boolean :=
     Controller.Register (Self_Pool'Access, Local_Partition_ID);

   procedure Test_Primarity
     (Pool     : access New_Pool_Type;
      Number   : in  Natural) is

      use Controller;
      use Ada.Text_IO;

   begin -- Test_Primarity

      --  adjust in case we received a new Number to test
      if Number_To_Test /= Number then
         Number_To_Test := Number;
         Current_Prime_Index := 1;
      end if;

      --  is there any other prime to test againt Number
      if Current_Prime_Index <= Last_Prime_Index then
         if Number mod Local_Prime_Table (Current_Prime_Index) = 0 then
            --  if Number is divisible by our current prime
            --+ then we just store the current prime in 'partition1'
            Results.Save
              (Local_Prime_Table (Current_Prime_Index),
               Local_Partition_ID);
         else
            --  otherwise Number could be a prime, so we charge
            --+ our neighbor partition to check it against its
            --+ set of primes by making an RPC to that partition
            --+
            --+ we may need to locate our neighbor partition
            --+ in which case we make a blocking call on Next
            --+ which gives us one
            if Next_Pool = null then
               Next_Pool := Controller.Next (Local_Partition_ID);
            end if;

            Current_Prime_Index := Current_Prime_Index + 1;

            --  dynamically-bound RPC
            Common.Test_Primarity (Next_Pool, Number);

         end if;
      else
         --  this branch captures our first prime
         --+ which we store in partition1
         Last_Prime_Index := Last_Prime_Index + 1;
         Local_Prime_Table (Last_Prime_Index) := Number;
         Results.Save
           (Local_Prime_Table (Current_Prime_Index),
            Local_Partition_ID);
      end if;

      --  we don't need all this verbosity, do we?
      --+-----------------------------------------

      --        Put ("I am partition " & Partition_ID'Image(Local_Partition_ID)
      --             & " and my prime table contains:");
      --        for I in 1 .. Last_Prime_Index loop
      --           Put (Natural'Image ( Local_Prime_Table (I)));
      --        end loop;
      --        Put_Line (" ***");
   end Test_Primarity;

end Prime;
